/* iTransactable v0.0.2
 * Copyright 2016, AnyWhichWay and Simon Y. Blackwell
 * Available under MIT license <https://mths.be/mit>
 */
(function() {
	"use strict";
	
	function iTransactable(constructorOrObject) {
		function makeTransactable(object) {
			Object.defineProperty(object,"beginTransaction",{configurable:true,writable:true,value:function(onError) {
				if(object.beginTransaction.transaction) {
					var err = new Error("Object already in transaction");
					if(onError) {
						onError(err);
					} else {
						throw err;
					}
				}
				var resolver,
					rejector,
					promise = new Promise((resolve,reject) => { resolver = resolve, rejector = reject; }),
					oldvalues = {},
					descriptors = {};
				object.beginTransaction.transaction = promise;
				Object.defineProperty(promise,"setTransactionValue",{configurable:true,writable:true,value:function(property,value) {
					if(oldvalues[property]===undefined) {
						oldvalues[property] = object[property];
					}
					if(object[property]===undefined) {
						descriptors[property] = {configurable:true,enumerable:true,get:function() { return descriptors[property].value; },set:function(value) { promise.setTransactionValue(key,value); }};
						Object.defineProperty(object,property,descriptors[property]);
					}
					descriptors[property].value = value;
				}});
				Object.keys(object).forEach((key) => {
					var desc = Object.getOwnPropertyDescriptor(object,key);
					if(desc.configurable && desc.writable) {
						descriptors[key] = desc;
						Object.defineProperty(object,key,{configurable:true,enumerable:true,get:function() { return desc.value; },
							set:function(value) { promise.setTransactionValue(key,value); }});
					}
				});
				Object.defineProperty(promise,"commit",{configurable:true,writable:true,value:function() {
					return promise.endTransaction(true);
				}});
				Object.defineProperty(promise,"endTransaction",{configurable:true,writable:true,value:function(committed=false) {
					Object.keys(descriptors).forEach((key) => {
						Object.defineProperty(object,key,descriptors[key]);
					});
					object.beginTransaction.transaction  = null;
					resolver(committed);
					return promise;
				}});
				Object.defineProperty(promise,"rollback",{configurable:true,writable:true,value: function(transactionError) {
					try {
						Object.keys(oldvalues).forEach(function(key) {
							if(oldvalues[key]===undefined) {
								delete object[key];
								delete descriptors[key];
							} else {
								object[key] = oldvalues[key];
							}
						});
						this.endTransaction();
						if(transactionError) {
							if(onError) {
								onError(transactionError);
							} else {
								rejector(transactionError);
							}
						}
					} catch(e) {
						var err = new Error("Unable to rollback transaction");
						err.transactionError = transactionError;
						err.rollbackError = e;
						if(onError) {
							onError(err);
						} else {
							rejector(err);
						}
					}
					return promise;
				}});
				return new Proxy(promise,{
					set: function(target,property,value) {
						if(target[property]===undefined) {
							object[property] = value;
						}
						target[property] = value;
						return true;
					}
				});
			}});
			return new Proxy(object,{
				set: function(target,property,value) {
					if(target.beginTransaction.transaction) {
						target.beginTransaction.transaction.setTransactionValue(property,value);
					} else {
						target[property] = value;
					}
					return true;
				}
			});
		}
		if(typeof(constructorOrObject)==="function") {
			return new Proxy(constructorOrObject,{
				construct: function(target,argumentsList) {
					var instance = new target(...argumentsList);
					return makeTransactable(instance);
				}
			});
		}
		return makeTransactable(constructorOrObject);
	}
	iTransactable.beginTransaction = function beginTransaction(objects,onError) {
		var resolver,
			rejector,
			promise = new Promise((resolve,reject) => { resolver = resolve, rejector = reject; }),
			targets = [];
		onError = (onError ? onError : function(err) { rejector(err); });
		objects.forEach(function(object) {
			if(typeof(object.beginTransaction)!=="function") {
				throw new Error("Object does not support transactions");
			}
			targets.push(object.beginTransaction(onError));
		});
		promise.commit = function() {
			targets.forEach(function(target) {
				target.commit();
			});
			return Promise.all(targets).then(function() {
				resolver(true);
			});
		}
		promise.rollback = function(err) {
			var promises = [];
			targets.forEach(function(target) {
				promises.push(target.rollback(err));
			});
			Promise.all(promises).then(function() {
				resolver(false);
			});
			return promise;
		}
		promise.end = function() {
			targets.forEach(function(target) {
				target.endTransaction(false);
			});
			return Promise.all(targets).then(function() {
				resolver(false);
			});
		}
		return promise;
	}
	
	if (typeof(module)!=="undefined") {
		module.exports  = iTransactable;
	}
	if(typeof(window)!=="undefined") {
		window.iTransactable = iTransactable;
	}
})();